/* -*-c-*-
 *
 * $Id: ec-raw.c,v 1.2 2004/04/08 01:36:15 mdw Exp $
 *
 * Raw formatting of elliptic curve points
 *
 * (c) 2004 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------*
 *
 * This file is part of Catacomb.
 *
 * Catacomb is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Catacomb is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with Catacomb; if not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

/*----- Header files ------------------------------------------------------*/

#include "ec.h"
#include "ec-raw.h"

/*----- Main code ---------------------------------------------------------*/

/* --- @ec_putraw@ --- *
 *
 * Arguments:	@ec_curve *c@ = elliptic curve
 *		@buf *b@ = pointer to a buffer
 *		@const ec *p@ = an elliptic curve point
 *
 * Returns:	Zero on success, nonzero on failure.
 *
 * Use:		Puts an elliptic curve point to the given buffer using the
 *		standard uncompressed format described in P1383 and SEC1.
 *		This requires at most @1 + 2 * c->f->noctets@ space in the
 *		buffer.  We don't do point compression.
 */

int ec_putraw(ec_curve *c, buf *b, const ec *p)
{
  octet *q;
  size_t n;

  if (EC_ATINF(p)) return (buf_putbyte(b, 0));
  buf_putbyte(b, 4);
  n = c->f->noctets;
  if ((q = buf_get(b, n * 2)) == 0) return (-1);
  mp_storeb(p->x, q, n);
  mp_storeb(p->y, q + n, n);
  return (0);
}

/* --- @ec_getraw@ --- *
 *
 * Arguments:	@ec_curve *c@ = elliptic curve
 *		@buf *b@ = pointer to a buffer
 *		@ec *d@ = an elliptic curve point
 *
 * Returns:	Zero on success, nonzero on failure.
 *
 * Use:		Reads an elliptic curve point from the given buffer using the
 *		standard uncompressed format described in P1383 and SEC1.
 *		We don't do point compression.
 */

int ec_getraw(ec_curve *c, buf *b, ec *d)
{
  const octet *q;
  size_t n;
  int u;

  if ((u = buf_getbyte(b)) < 0) return (-1);
  if (!u) { EC_SETINF(d); return (0); }
  if (!(u & 4)) return (-1);
  n = c->f->noctets;
  if ((q = buf_get(b, n * 2)) == 0) return (-1);
  EC_DESTROY(d);
  d->x = mp_loadb(MP_NEW, q, n);
  d->y = mp_loadb(MP_NEW, q + n, n);
  d->z = 0;
  return (0);
}

/*----- That's all, folks -------------------------------------------------*/
